﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;

namespace AT{
    namespace UBF{
        namespace Behaviors {
            /** IF bullets are in view, determine if they're in one of three locations:
             * current position, behind the tank, in front of the tank.  Move to which
             * ever option has the least incoming bullets.
             *
             */
            public class GoodAgent_Defense : AT.UBF.Behavior {
                private float buffer = 2;
                public float priority = 0.8f;

                protected override AT.UBF.Action genActionInternal(AT.UBF.State state){
					AT.UBF.Action action = new AT.UBF.Action();
                    
					int visibleECount = state.visibleEnemyTanks.Count;
					int visibleBCount = state.visibleBullets.Count;
					int i = 0;
					bool visibleEnemy = false;
					Vector2 targetPos = new Vector2(0,0);  
                    
					//if enemy in sight, set priority and get enemy position
					while(i < visibleECount && visibleEnemy == false){
						AT.Team targetTeam = state.visibleEnemyTanks [i].getTeam();
						if(targetTeam != state.self.getTeam()){					
							action.SetPriority(priority, voteSourceID);
							visibleEnemy = true;				
							targetPos = state.visibleEnemyTanks[i].getPosition();  
						}
						i++;
                    }
					
					//turn perpendicular to enemy
					if(visibleEnemy){
						action.SetTurnTracksRight(findShortestTrackRotationToPerpTarget(state.self.getPosition(), targetPos, state.self.getTracksRotation()), voteSourceID);
					}

					//move away from closest two bullets
					if(visibleBCount > 0)
					{
						action.SetPriority(priority, voteSourceID);
						Triple<double, int, int>  bullets = closestBullets (state);
						if (bullets.first < 12)
						if (bullets.second == 0) {
							if (bullets.third == 1)
								action.SetDriveForward (-1.0f, voteSourceID);
							else
								action.SetDriveForward (1.0f, voteSourceID); 
						} else if (bullets.second == 1) {
							if(bullets.third == -1)
								action.SetDriveForward(0.0f, voteSourceID);
							else
								action.SetDriveForward (-1.0f, voteSourceID);

						} if (bullets.second == -1) {
							if (bullets.third == 1)
								action.SetDriveForward (0.0f, voteSourceID);
							else
								action.SetDriveForward (1.0f, voteSourceID);
						}
						
					}
					
					
					return action;
				}

				Triple<double, int, int> closestBullets(AT.UBF.State state){
					Vector2 myPosition = state.self.getPosition();
					Vector2 myVelocity = state.self.getVelocity();
					Vector2 tankBuffer = state.bodyLength.first - state.bodyLength.second;
					double d1 = 200;
					double d2 = 200;
					int s1 = 0;
					int s2 = 0;
				
					Vector2 futureTankPos = myPosition + myVelocity + tankBuffer;
					Vector2 pastTankPos = myPosition - myVelocity - tankBuffer;
					float dirAngle = Util.getRelativeAngleToTarget(myPosition, futureTankPos);
					float relAngle = dirAngle - state.self.getTracksRotation();

					Tuple<Vector2, Vector2> futureLine = new Tuple<Vector2, Vector2>(myPosition, futureTankPos);
					Tuple<Vector2, Vector2> pastLine = new Tuple<Vector2, Vector2>(myPosition, pastTankPos);

					foreach(Tuple<Vector2, Vector2> bullet in state.visibleBullets){
						Tuple<Vector2, Vector2> currTankLine = getMaxTankWidthLine(bullet, state.bodyLength, myPosition);
						Vector2 bulletPos = bullet.first;
						Vector2 bulletVel = bullet.second;


						Vector2 bulletDest = bulletPos + bulletVel * buffer;
						Tuple<Vector2, Vector2> bulletLine = new Tuple<Vector2, Vector2>(bulletPos, bulletDest);

						if(LineIntersection.doLinesIntersect(currTankLine, bulletLine)){
							double dist = Util.distToLineSegment (myPosition, bulletPos, bulletDest).first;
							if (dist < d1) {
								d2 = d1;
								s2 = s1;
								d1 = dist;
								s1 = 0;
							} else if (dist < d2) {
								d2 = dist;
								s2 = 0;
							}

						}
						else if(LineIntersection.doLinesIntersect(futureLine, bulletLine)){
							double dist = Util.distToLineSegment (myPosition, bulletPos, bulletDest).first;
							if (dist < d1) {
								d2 = d1;
								s2 = s1;
								d1 = dist;
								s1 = 1;
							} else if (dist < d2) {
								d2 = dist;
								s2 = 1;
							}
						}
						else if(LineIntersection.doLinesIntersect(pastLine, bulletLine)){
							double dist = Util.distToLineSegment (myPosition, bulletPos, bulletDest).first;
							if (dist < d1) {
								d2 = d1;
								s2 = s1;
								d1 = dist;
								s1 = -1;
							} else if (dist < d2) {
								d2 = dist;
								s2 = -1;
							}
						}
					}
					Triple<double, int, int> closestBullets = new Triple<double, int, int>(d1,s1,s2);

					return closestBullets;

				}

				float findShortestTrackRotationToPerpTarget(Vector2 myPos, Vector2 targetPos, float myTrackRotation){
					float targetRotation = Util.getRelativeAngleToTarget(myPos, targetPos); 
					targetRotation = targetRotation - 90f;
					float maxRot = AT.Tank.trackRotationSpeed;
					return findShortestRotation(targetRotation, myTrackRotation, maxRot);
				} 

				float findShortestRotation(float targetRotation, float currRotation, float maxRotation){
					float temp = (targetRotation - currRotation + 360) % 360;
					float turnRight = 0.0f;

					if (temp > 180) {
						if(temp > 360-maxRotation){ 
							turnRight = Mathf.Abs(360-temp)/maxRotation;
						}else{
							turnRight = 1.0f;
						}
					}else{
						if(temp < maxRotation){
							turnRight = (0-temp)/maxRotation;
						}else{
							turnRight = -1.0f;
						}
					}

					return turnRight;
				}

				
                private Tuple<Vector2, Vector2> getMaxTankWidthLine(Tuple<Vector2, Vector2> bulletLine, Tuple<Vector2, Vector2> bodyLength, Vector2 tankPos) {
                    Vector2 v = bulletLine.second;
                    Vector2 numerator = new Vector2(-v.y, v.x);

                    Vector2 tankSize = (bodyLength.first - bodyLength.second) * 1.2f;
                    float scaling = Mathf.Sqrt(Mathf.Pow(tankSize.x, 2) + Mathf.Pow(tankSize.y, 2)) / 2f;

                    Vector2 p3 = tankPos + numerator / Mathf.Sqrt(Mathf.Pow(v.x, 2) + Mathf.Pow(v.y, 2)) * scaling;
                    Vector2 p4 = tankPos + numerator / Mathf.Sqrt(Mathf.Pow(v.x, 2) + Mathf.Pow(v.y, 2)) * -scaling;
                    
                    return new Tuple<Vector2, Vector2>(p3, p4);
                }
                
                public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction){
                    dict["Priority"] = priority;
                    dict["Buffer"] = buffer;
                }
                
                public override void _FromDict(Dictionary<string, object> dict){
                    string bName = this.GetType().Name;
                    priority =  Util.FloatFromJsonDict("Priority", priority, bName, dict);
                    buffer = Util.FloatFromJsonDict("Buffer", buffer, bName, dict);
                }
            }
        }
    }
}