﻿using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;

namespace AT{
    namespace UBF{
        namespace Behaviors{
            /** Turns turret until we are looking directly at an enemy. Once one is in sight, try to point the turret directly at them (pure pursuit).
             *
             *  Does **not** vote on rotating the tracks, driving or firing the gun.
             */
            public class GoodAgent_FindEnemy : AT.UBF.Behavior {
                public float priorityWhenNoEnemyInSight = 1.0f; // highest priority
                public float priorityWhenEnemyInSight   = 0.1f;


                protected override AT.UBF.Action genActionInternal(AT.UBF.State state){
                    AT.UBF.Action action = new AT.UBF.Action();

					int visibleCount = state.visibleEnemyTanks.Count;
					int i = 0;
					bool visibleEnemy = false;
					float turretRotateSpeed = 1.0f; // rotate at full speed
					float priority = priorityWhenNoEnemyInSight;
                    
					//changed to prevent targeting teammate
					while(i < visibleCount && visibleEnemy == false){
						AT.Team targetTeam = state.visibleEnemyTanks [i].getTeam();
						if(targetTeam != state.self.getTeam()){
							Vector2 targetPos = state.visibleEnemyTanks[i].getPosition();        
	                        //Debug.DrawLine(state.position, targetPos,  Color.red, 1f);
	                        // Pure pursuit (no lag or lead on turret)
							turretRotateSpeed = Util.findShortestTurretRotationToTarget(state.self.getPosition(), targetPos, state.self.getTurretRotation());					
							priority = priorityWhenEnemyInSight;
							visibleEnemy = true;
						}
						i++;
                    }

					action.SetTurnTurretRight(turretRotateSpeed, voteSourceID);
					action.SetPriority(priority, voteSourceID);

                    return action;
                }

                /** Save arguments to dictionary.
                 */
                public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction){
                    dict["PriorityWhenNoEnemyInSight"] = priorityWhenNoEnemyInSight;
                    dict["PriorityWhenEnemyInSight"] = priorityWhenEnemyInSight;
                }

                /** Loads arguments from a dictionary, verifying as needed.
                 */
                public override void _FromDict(Dictionary<string, object> dict){
                    string bName = this.GetType().Name;
                    priorityWhenNoEnemyInSight = Util.FloatFromJsonDict("PriorityWhenNoEnemyInSight", priorityWhenNoEnemyInSight, bName, dict);
                    priorityWhenEnemyInSight = Util.FloatFromJsonDict("PriorityWhenEnemyInSight", priorityWhenEnemyInSight, bName, dict);
                }
            }
        }
    }
}