using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;

namespace AT{
    namespace UBF{
        namespace Behaviors{
            /** 
             * Wanders and does not get stuck in corners
            */

			//TODO handle bullet situation

            class GoodAgent_Wandering : AT.UBF.Behavior{
                public float crashBuffer = 2.0f; //higher means turn earlier
				public float priorityCollisionEminent = 0.9f;
				public float priorityStuck = 0.9f;
				public float priorityOtherwise = 0.1f;

                protected override AT.UBF.Action genActionInternal(AT.UBF.State state) {
					AT.UBF.Action action = new AT.UBF.Action ();
				
					float speed = 1.0f;
					
					if(state.visibleBullets.Count < 1)
						action.SetDriveForward (speed, voteSourceID);

					//position of front, left and right sides of tank
					Vector2 tankFront = state.bodyLength.first;
					Vector2 tankLeft = state.bodyWidth.first; 
					Vector2 tankRight = state.bodyWidth.second;

					//velocity data
					Vector2 velMagnitude = state.self.getVelocity() * crashBuffer;
					Vector2 velocityLeftSide = tankLeft + velMagnitude;
					Vector2 velocityRightSide = tankRight + velMagnitude;
					Tuple<Vector2, Vector2> velocityLine1 = new Tuple<Vector2, Vector2>(tankLeft, velocityLeftSide);
					Tuple<Vector2, Vector2> velocityLine2 = new Tuple<Vector2, Vector2>(tankRight, velocityRightSide);

					//check if left and right sides will collide with closest wall
					Tuple<Vector2,Vector2> wall = findClosestWall (tankFront, state);
					double distance = Util.distToLineSegment (tankFront, wall.first, wall.second).first;
					bool leftCrash = LineIntersection.doLinesIntersect (velocityLine1, wall);
					bool rightCrash = LineIntersection.doLinesIntersect (velocityLine2, wall);
					bool crashTest = leftCrash || rightCrash;
					bool collision = (distance < 1);
					bool stuck = (state.self.getVelocity ().x == 0 && state.self.getVelocity ().y == 0);
					float priority = priorityOtherwise;

					//if crash emminent or stuck (i.e., velocity = 0)
					if (crashTest || stuck || collision) {
						//if the closest two walls make a corner, get out!
						if (isCorner (state.bodyLength.first, state)|| stuck || collision) {
							action.SetTurnTracksRight (-0.5f, voteSourceID);
							speed = 0.5f;
							priority = priorityStuck;
						}
						//otherwise check right and left crash values
						else {
							priority = priorityCollisionEminent;
							//if both are true, check which side is closest to the wall and turn appropriately
							if (leftCrash && rightCrash) {
								double distanceL = Util.distToLineSegment (tankLeft, wall.first, wall.second).first;
								double distanceR = Util.distToLineSegment (tankRight, wall.first, wall.second).first;
								if(distanceR < distanceL)
									action.SetTurnTracksRight (-1.0f, voteSourceID);
								else
									action.SetTurnTracksRight (1.0f, voteSourceID);
								speed = 0.5f;
							}
							//otherwise, if the leftside is going to crash turn appropriately
							else if (leftCrash) {
									action.SetTurnTracksRight (1.0f, voteSourceID);
									speed = 0.5f;
							}
							//same with rightside
							else if (rightCrash) {
								action.SetTurnTracksRight (-1.0f, voteSourceID);
								speed = 0.5f;
							}
						}
						action.SetDriveForward (speed, voteSourceID);
					}	
					action.SetPriority(priority, voteSourceID);


					return action;

				}
  
				//check if two closest walls make a corner
				bool isCorner(Vector2 self, AT.UBF.State state){

					Tuple<Vector2,Vector2> closestWall = state.walls[1];
					Tuple<Vector2,Vector2> secondClosestWall = state.walls[1];
					double closest = 200;
					double secondClosest = 200;

					foreach (Tuple<Vector2, Vector2> wall in state.walls){
						double dist = Util.distToLineSegment (self, wall.first, wall.second).first;
						if (dist < closest) {
							secondClosest = closest;
							secondClosestWall = closestWall;
							closest = dist;
							closestWall = wall;

						} else if (dist < secondClosest) {
							secondClosest = dist;
							secondClosestWall = wall;
						}

					}

					if (LineIntersection.doLinesIntersect (closestWall, secondClosestWall))
						return true;
					else
						return false;
				}

				//find the closest wall
				Tuple<Vector2,Vector2> findClosestWall(Vector2 self, AT.UBF.State state){
					
					Tuple<Vector2,Vector2> closestWall = state.walls[1];
					double closest = 200;

					foreach (Tuple<Vector2, Vector2> wall in state.walls){
	
						double dist = Util.distToLineSegment (self, wall.first, wall.second).first;
						if (dist < closest) {
							closest = dist;
							closestWall = wall;
						}

					}
					return closestWall;
				}

                public override void _ToDict(Dictionary<string, object> dict, bool saveLastAction){
					dict["PriorityCollisionEminent"] = priorityCollisionEminent;
					dict["PriorityOtherwise"] = priorityOtherwise;
					dict["PriorityStuck"] = priorityStuck;
                }
                
                
                public override void _FromDict(Dictionary<string, object> dict){
                    string bName = this.GetType().Name;
					priorityCollisionEminent = Util.FloatFromJsonDict("PriorityCollisionEminent", priorityCollisionEminent, bName, dict);
					priorityOtherwise = Util.FloatFromJsonDict("PriorityOtherwise", priorityOtherwise, bName, dict);
					priorityStuck = Util.FloatFromJsonDict("PriorityStuck", priorityStuck, bName, dict);
                }
            }
        }
    }
}